import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)
import java.util.List;
/**
 * Mouse tool which serves as the primary means of adding and removing Objects in the Field.
 * Left mouse button places the selected Object and right mouse button removes all Objects in the Cursor.
 * The size of the Cursor can be changed using the up and down arrow keys, larger and smaller respectively.
 * 
 * @author Mr. Hedgeapple 
 * @version 3.0
 */
public class Cursor extends Actor
{
    //Default Brush Size
    int bSize = 50/Field.size;
    
    //Mouse button booleans
    boolean mB1 = false;
    boolean mB2 = false;
    
    /**
     * Basic constructor for Cursor.
     */
    public Cursor()
    {
        update();
    }

    /**
     * Act - do whatever the Cursor wants to do. This method is called whenever
     * the 'Act' or 'Run' button gets pressed in the environment.
     */
    public void act() 
    {
        update();
        track();
        brushSize();
        place();
        remove();
    }    

    /**
     * Updates the image of Cursor.
     */
    public void update()
    {
        int width = bSize*Field.size;
        setImage(new GreenfootImage(width, width));
        GreenfootImage img = getImage();
        img.setColor(Color.WHITE);
        img.drawRect(0, 0, width-1, width-1);
    }

    /**
     * Makes the Cursor follow the mouse.
     */
    public void track()
    {
        World myWorld = getWorld();
        MouseInfo mouse = Greenfoot.getMouseInfo();
        if (mouse != null && mouse.getX() < myWorld.getWidth()-(100/Field.size))
        {
            setLocation(mouse.getX(), mouse.getY());
            mouseButton();
        } else {
            setLocation(450/Field.size, 300/Field.size);
            mB1 = false;
            mB2 = false;
        }
    }

    /**
     * Change the Brush Size using up and down arrow keys.
     */
    public void brushSize()
    {
        if (Greenfoot.isKeyDown("up") && bSize < 150/Field.size)
            bSize++;
        if (Greenfoot.isKeyDown("down") && bSize > 1)
            bSize--;
    }

    /**
     * Detects mouse input.
     */
    public void mouseButton()
    {
        MouseInfo mouse = Greenfoot.getMouseInfo();
        if (Greenfoot.mousePressed(null) && mouse.getButton() == 1)
            mB1 = true;
        else if (Greenfoot.mouseClicked(null) && mouse.getButton() == 1)
            mB1 = false;
        if (Greenfoot.mousePressed(null) && mouse.getButton() == 3)
            mB2 = true;
        else if (Greenfoot.mouseClicked(null) && mouse.getButton() == 3)
            mB2 = false;
    }
    
    /**
     * Places Objects of selected type in the Field within the Cursor's range.
     */
    public void place()
    {
        World myWorld = getWorld();
        int range = bSize/2;
        
        if (mB1)
        {
            for (int x = getX()-range; x <= getX()+range; x++)
            {
                for (int y = getY()-range; y <= getY()+range; y++)
                {
                    if (x >= 0 && x < myWorld.getWidth()-(100/Field.size) && y >= 0 && y < myWorld.getHeight())
                    {
                        Actor a = getOneObjectAtOffset(x-getX(), y-getY(), Objects.class);
                        if (a == null)
                            myWorld.addObject(new Objects(Field.obj), x, y);
                    }
                }
            }
        }
    }
    
    /**
     * Removes all Objects within the Cursor's range after right-click.
     */
    public void remove()
    {
        World myWorld = getWorld();
        int range = bSize/2;
        
        if (mB2)
        {
            for (int x = getX()-range; x <= getX()+range; x++)
            {
                for (int y = getY()-range; y <= getY()+range; y++)
                {
                    if (x >= 0 && x < myWorld.getWidth()-(100/Field.size) && y >= 0 && y < myWorld.getHeight())
                    {
                        Actor a = getOneObjectAtOffset(x-getX(), y-getY(), Objects.class);
                        if (a != null)
                            myWorld.removeObject(a);
                    }
                }
            }
        }
    }
}
