import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)
import java.util.List;
/**
 * The behavior of the Object is dependent on its type. Many interactions are possible between Objects.
 * 
 * @author Mr. Hedgeapple 
 * @version 3.6
 */
public class Objects extends Actor
{
    //Constants
    final int roomTemp = 21;

    //Class definitions
    public int type;
    public int getType()
    {
        return type;
    }
    public String name;
    public double heat;
    public double getHeat()
    {
        return heat;
    }

    public void setHeat(int temp)
    {
        heat = temp;
    }
    int r, g, b;
    //Values for heat transfer
    public double k;
    public double c;
    public double m;
    public int state;
    public int getState()
    {
        return state;
    }

    //Definition arrays
    String names[] = 
        {"Wall", "Water", "Stone", "Fire", "Clone", "Lava", "Metal", "Vine", "Glass", "Powder", "Spark"};
    int ri[] = {125, 0,   100, 255, 100, 255, 50, 0,   0, 242, 255};
    int gi[] = {125, 0,   100, 0,   15,  170,  50, 255, 0, 227, 240};
    int bi[] = {125, 255, 100, 0,   15,  15,  50, 0,   0, 118, 30};
    int heats[] =
        {roomTemp, roomTemp, roomTemp, 1600, roomTemp, 1200, roomTemp, roomTemp, roomTemp, roomTemp, 260};
    double thermalConductivity[] =
        {0, 0.58, 4.5, 0, 0, 0.5, 80, 0.23, 1.05, 0.08, 0};
    //J/g*Cel
    double specificHeat[] =
        {0, 4.182, 0.41, 0, 0, 1.5, 0.449, 1.4, 0.67, 2.4, 0};
    //g for 1cm^3
    double mass[] =
        {1, 1, 2.4, 0.0003, 1, 3.1, 7.874, 1, 2.579, .5, 0.0003};
    //Initial physical state of an Object.
    int iniState[] =
        {0, 1, 0, 2, 0, 1, 0, 0, 0, 0, 2};
    //Melting points of various Objects. Objects which do not transfer heat have melting point of roomTemp.
    int meltingPoint[] =
        {roomTemp, 0, 540, roomTemp, roomTemp, 540, 1538, roomTemp, 1400, roomTemp, roomTemp};

    //Movement ints
    public int dx;
    public int getDx()
    {
        return dx;
    }
    public int dy;
    public int getDy()
    {
        return dy;
    }

    /**
     * Basic constructor for Objects.
     */
    public Objects(int classType)
    {
        type = classType;
        name = names[type];
        heat = heats[type];
        state = iniState[type];
        k = thermalConductivity[type];
        c = specificHeat[type];
        m = mass[type];
        r = ri[type];
        g = gi[type];
        b = bi[type];
        setImage(new GreenfootImage(Field.size, Field.size));
        GreenfootImage img = getImage();
        img.setColor(new Color(r, g, b));
        img.fill();
        //Object type Spark will always be electrified.
        if (type == 10)
            electrified = true;
    }

    /**
     * Act - do whatever the Objects wants to do. This method is called whenever
     * the 'Act' or 'Run' button gets pressed in the environment.
     */
    public void act() 
    {
        //Wall behavior. Does not react with other Objects and only serves as a barrier.
        /**if (type == 0)
        {
        Actor o = getOneIntersectingObject(Objects.class);
        if (o != null)
        getWorld().removeObject(o);
        }**/

        if (!PauseButton.paused)
        {
            switch (type) {
                case 1: waterBehavior(); break;
                case 2: stoneBehavior(); break;
                case 3: fireBehavior(); break;
                case 4: cloneBehavior(); break;
                case 5: lavaBehavior(); break;
                case 6: metalBehavior(); break;
                case 7: vineBehavior(); break;
                case 8: glassBehavior(); break;
                case 9: powderBehavior(); break;
                case 10: sparkBehavior(); break;
            }
        }
        else if (type == 8 && !electrified) 
        //Method specific to Glass. Will draw its border even when simulation is paused.
            drawBorder();
        //formGroups();
    }    
    //Boolean denotes whether the Object is currently carrying an electrical current.
    public boolean electrified = false;
    public boolean getElectrified()
    {
        return electrified;
    }

    /**
     * Behavior for type Water. At room temperature will flow like a liquid. Will turn into steam upon
     * reaching its boiling point and will turn into ice upon reaching its freezing point. Can quickly
     * absorb large amounts of heat off of other Objects.
     */
    private void waterBehavior()
    {
        transfer();
        if (state == 1)
            movementLqd();
        else if (state == 2)
            movementGas();
        //if (state == 1)
        //electrify();
        if (!electrified)
            waterStateChange();
    }

    /**
     * Changes the appearance of Water based on its current temperature.
     */
    private void waterStateChange()
    {
        if (state == 0 && getImage().getColor() != new Color(170, 200, 255))
        {
            GreenfootImage solid = new GreenfootImage(Field.size, Field.size);
            solid.setColor(new Color(170, 200, 255));
            solid.fill();
            setImage(solid);
        }
        else if (state == 1 && getImage().getColor() != Color.BLUE)
        {
            GreenfootImage liquid = new GreenfootImage(Field.size, Field.size);
            liquid.setColor(Color.BLUE);
            liquid.fill();
            setImage(liquid);
        }
        else if (state == 2 && getImage().getColor() != new Color(220, 235, 255))
        {
            GreenfootImage gas = new GreenfootImage(Field.size, Field.size);
            gas.setColor(new Color(220, 235, 255));
            gas.fill();
            setImage(gas);
        }
    }

    /**
     * Behavior for type Stone. Affected by gravity and forms piles on solid ground. Will displace liquids
     * such as Water and Lava. Can be heated to the point of melting and will then function as Lava.
     */
    private void stoneBehavior()
    {
        transfer();
        if (state == 0)
            movementStone();
        else
            stoneLavaChange();
    }

    /**
     * Movement for type Stone.
     */
    private void movementStone()
    {
        World myWorld = getWorld();
        Actor d = getOneObjectAtOffset(0, 1, Objects.class);
        Actor dl = getOneObjectAtOffset(-1, 1, Objects.class);
        Actor dr = getOneObjectAtOffset(1, 1, Objects.class);

        if (getY() < myWorld.getHeight()-1)
        {
            if (d == null)
                dy = 1;
            else if (((Objects)(d)).getType() == type && ((Objects)(d)).getDy() == 1)
                dy = 1;
            else if (((Objects)(d)).getState() == 1)
            {
                dy = 1;
                ((Objects)(d)).setLocation(getX(), getY());
            } else {
                if (Greenfoot.getRandomNumber(2) == 0)
                {
                    if (dl == null)
                    {
                        dy = 1;
                        dx = -1;
                    }
                    else if (((Objects)(dl)).getState() == 1)
                    {
                        dy = 1;
                        dx = -1;
                        ((Objects)(dl)).setLocation(getX(), getY());
                    }
                    else if (dr == null)
                    {
                        dy = 1;
                        dx = 1;
                    }
                    else if (((Objects)(dr)).getState() == 1)
                    {
                        dy = 1;
                        dx = 1;
                        ((Objects)(dr)).setLocation(getX(), getY());
                    } else
                        return;
                } else {
                    if (dr == null)
                    {
                        dy = 1;
                        dx = 1;
                    }
                    else if (((Objects)(dr)).getState() == 1)
                    {
                        dy = 1;
                        dx = 1;
                        ((Objects)(dr)).setLocation(getX(), getY());
                    }
                    else if (dl == null)
                    {
                        dy = 1;
                        dx = -1;
                    }
                    else if (((Objects)(dl)).getState() == 1)
                    {
                        dy = 1;
                        dx = -1;
                        ((Objects)(dl)).setLocation(getX(), getY());
                    } else
                        return;
                }
            }
        } else
            return;

        int x = getX()+dx;
        if (x >= 0 && x < myWorld.getWidth()-(100/Field.size))
            movement();
        dx = 0;
        dy = 0;
    }

    int lifetime = Greenfoot.getRandomNumber(100);
    /**
     * Behavior for type Fire. Will rapidly rise through the air. Dissipates after a random amount of time.
     * Will heat and set fire to certain Objects.
     */
    private void fireBehavior()
    {
        movementGas();
        if (lifetime <= 0)
            getWorld().removeObject(this);
        else
            lifetime--;
    }

    /**
     * Behavior for type Clone. Will continuously replicate the first Object it touches (excluding Clone and Wall).
     */
    private void cloneBehavior()
    {
        if (!objectSet)
            findObject();
        else
            spawnObject();
    }
    boolean objectSet = false;
    int objectType = 0;
    /**
     * Searches for adjacent Objects. 
     * It will continuously spawn the first type of Objects it touches, excluding Clone and Wall.
     */
    private void findObject()
    {
        if (!getNeighbours(1, true, Objects.class).isEmpty())
        {
            List objects = getNeighbours(1, true, Objects.class);
            for (int i = 0; i < objects.size(); i++)
            {
                if (((Objects)(objects.get(i))).getType() != type && ((Objects)(objects.get(i))).getType() != 0)
                {
                    objectSet = true;
                    objectType = ((Objects)(objects.get(i))).getType();
                }
            }
        }
    }

    /**
     * Spawns Objects in random empty cells adjacent to the Clone.
     */
    private void spawnObject()
    {
        World myWorld = getWorld();

        for (int x = getX()-1; x <= getX()+1; x++)
        {
            if (x >= 0 && x < myWorld.getWidth()-(100/Field.size))
            {
                for (int y = getY()-1; y <= getY()+1; y++)
                {
                    if (y >= 0 && y < myWorld.getHeight())
                    {
                        Actor object = getOneObjectAtOffset(x-getX(), y-getY(), Objects.class);
                        if (Greenfoot.getRandomNumber(30) == 0 && object == null && (x != getX() && y != getY()))
                            myWorld.addObject(new Objects(objectType), x, y);
                    }
                }
            }
        }
    }

    /**
     * Behavior for type Lava. Will flow like a liquid. At a certain temperature it will turn into Stone.
     * Will heat and set fire to certain Objects.
     */
    private void lavaBehavior()
    {
        transfer();
        if (state == 1)
            movementLqd();
        else
            stoneLavaChange();
    }

    /**
     * Controls the physical state change between Stone and Lava.
     */
    private void stoneLavaChange()
    {
        GreenfootImage stone = new GreenfootImage(Field.size, Field.size);
        stone.setColor(new Color(100, 100, 100));
        stone.fill();

        GreenfootImage lava = new GreenfootImage(Field.size, Field.size);
        lava.setColor(new Color(255, 170, 15));
        lava.fill();
        if (state == 0 && getImage() != stone)
        {
            type = 2;
            name = names[type];
            setImage(stone);
        }
        else if (state == 1 && getImage() != lava)
        {
            type = 5;
            name = names[type];
            setImage(lava);
        }
        k = thermalConductivity[type];
        c = specificHeat[type];
        m = mass[type];
    }

    /**
     * Behavior for type Metal. Can be heated and will appear to glow with heat. Has a high melting point
     * and will move like a liquid when that point is reached.
     */
    private void metalBehavior()
    {
        transfer();
        electrify();
        if (!electrified)
            glow();
        if (state == 1)
            movementLqd();
    }

    /**
     * Creates visual effect of heat glow for Metal.
     */
    private void glow()
    {   
        if (heat <= 21)
        {
            r = 50;
            g = 50;
            b = 50;
        } 
        else if (heat <= 199)
        {
            int decrease = (int)((heat - 21) / 3.569);
            g = 50 - decrease;
            b = 50 - decrease;
        }
        else if (heat <= 931)
        {
            int increase = (int)((heat - 199) / 3.569);
            r = 50 + increase;
            g = 0;
            b = 0;
        }
        else if (heat <= 1538)
        {
            g = (int)((heat - 931) / 3.569);
            r = 255;
            b = 0;
        }
        else
        {
            r = 255;
            g = 170;
            b = 0;
        }

        r = (r <= 255) ? r: 255;
        g = (g >= 0) ? g: 0;
        b = (b >= 0) ? b: 0;
        if (getImage().getColor() != new Color(r, g, b))
        {
            getImage().setColor(new Color(r, g, b));
            getImage().fill();
        }
    }

    /**
     * Behavior for type Vine. Will climb over adjacent Objects (excluding Vine) and drop tendrils of random
     * length. Will catch fire when in contact with Objects of a certain heat value.
     */
    private void vineBehavior()
    {
        burn();
        if (alive)
            tendril();
        spread();
    }
    //Controls whether the Vine is still able to drop tendrils.
    boolean alive = true;
    //Denotes whether the Vine is dropping a tendril.
    boolean dropping = false;
    //Limits the amount of time a Vine can remain alive.
    int timeAlive = 500;
    /**
     * Randomly drops tendrils.
     */
    private void tendril()
    {
        World myWorld = getWorld();
        Actor d = getOneObjectAtOffset(0, 1, Objects.class);

        if (d != null)
            alive = false;

        if (Greenfoot.getRandomNumber(4) == 0 || dropping)
        {
            //Above number controls the likelihood the Vine will drop any tendrils.
            dropping = true;
            if (Greenfoot.getRandomNumber(50) == 0)
            {
                //Above number controls the frequency at which the tendril will lengthen.
                setLocation(getX(), getY()+1);
                myWorld.addObject(new Objects(7), getX(), getY()-1);
                if (Greenfoot.getRandomNumber(10) == 0 || getY() == myWorld.getHeight()-1)
                    alive = false;
            }
        } else
            alive = false;

        if (timeAlive > 0)
            timeAlive--;
        else
            alive = false;
    }

    /**
     * Allows Vine to grow around other Objects excluding other Vines.
     */
    private void spread()
    {
        World myWorld = getWorld();

        if (!getNeighbours(1, true, Objects.class).isEmpty())
        {
            List obj = getNeighbours(1, true, Objects.class);
            for (int i = 0; i < obj.size(); i++)
            {
                int xOffset = 0;
                int yOffset = 0;
                if (((Objects)(obj.get(i))).getType() != 7 && Greenfoot.getRandomNumber(30) == 0)
                {
                    xOffset = ((Objects)(obj.get(i))).getX() - getX();
                    yOffset = ((Objects)(obj.get(i))).getY() - getY();

                    if (xOffset != 0 && yOffset != 0)
                    {
                        if (Greenfoot.getRandomNumber(2) == 0)
                        {
                            if (getOneObjectAtOffset(xOffset, 0, Objects.class) == null)
                                myWorld.addObject(new Objects(7), getX() + xOffset, getY());
                        } 
                        else if (getOneObjectAtOffset(0, yOffset, Objects.class) == null)
                            myWorld.addObject(new Objects(7), getX(), getY() + yOffset);
                    }
                    else if (xOffset != 0)
                    {
                        if (Greenfoot.getRandomNumber(2) == 0)
                        {
                            if (getOneObjectAtOffset(0, 1, Objects.class) == null)
                                myWorld.addObject(new Objects(7), getX(), getY() + 1);
                        }
                        else if (getOneObjectAtOffset(0, -1, Objects.class) == null)
                            myWorld.addObject(new Objects(7), getX(), getY() - 1);
                    }
                    else if (yOffset != 0)
                    {
                        if (Greenfoot.getRandomNumber(2) == 0)
                        {
                            if (getOneObjectAtOffset(1, 0, Objects.class) == null)
                                myWorld.addObject(new Objects(7), getX() + 1, getY());
                        }
                        else if (getOneObjectAtOffset(-1, 0, Objects.class) == null)
                            myWorld.addObject(new Objects(7), getX() - 1, getY());
                    }
                }
            }
        }
    }

    /**
     * Behavior for type Glass. Draws a border around contiguous sections of Glass. Can be heated and will
     * move like a liquid when it reaches its melting point.
     */
    private void glassBehavior()
    {
        transfer();
        electrify();
        if (!electrified)
            drawBorder();
        if (state == 1)
            movementLqd();
    }
    //Reference int which helps prevent unnecessary image updating in order to reduce lag.
    int adjacent = 0;
    /**
     * Draws a white border around the edges of the image not touching other Glass.
     */
    private void drawBorder()
    {
        //Image parts
        GreenfootImage fullBorder = new GreenfootImage(Field.size, Field.size);
        fullBorder.setColor(Color.WHITE);
        fullBorder.drawRect(0, 0, Field.size-1, Field.size-1);

        GreenfootImage vertical = new GreenfootImage(1, Field.size);
        vertical.setColor(Color.WHITE);
        vertical.drawLine(0, 0, 0, Field.size);

        GreenfootImage horizontal = new GreenfootImage(Field.size, 1);
        horizontal.setColor(Color.WHITE);
        horizontal.drawLine(0, 0, Field.size, 0);

        //Finding adjacent Glass and setting the image.
        if (!getObjectsInRange(1, Objects.class).isEmpty())
        {
            int adjacentTemp = 0;
            boolean u = false;
            boolean d = false;
            boolean l = false;
            boolean r = false;
            List obj = getObjectsInRange(1, Objects.class);
            for (int i = 0; i < obj.size(); i++)
            {
                if (((Objects)(obj.get(i))).getType() == type)
                {
                    adjacentTemp++;
                    if (((Objects)(obj.get(i))).getY() == getY()-1) u = true;
                    if (((Objects)(obj.get(i))).getY() == getY()+1) d = true;
                    if (((Objects)(obj.get(i))).getX() == getX()-1) l = true;
                    if (((Objects)(obj.get(i))).getX() == getX()+1) r = true;
                }
            }

            if (adjacentTemp != adjacent)
            {
                adjacent = adjacentTemp;
                if (!u && !d && !l && !r)
                    setImage(fullBorder);
                else {
                    GreenfootImage img = new GreenfootImage(Field.size, Field.size);
                    if (!u) img.drawImage(horizontal, 0, 0);
                    if (!d) img.drawImage(horizontal, 0, Field.size-1);
                    if (!l) img.drawImage(vertical, 0, 0);
                    if (!r) img.drawImage(vertical, Field.size-1, 0);
                    setImage(img);
                }
            }
        } else if (getImage() != fullBorder)
            setImage(fullBorder);
    }

    /**
     * Behavior for type Powder. Will float downward slowly. Will catch fire when in contact with Objects
     * of a certain heat value.
     */
    private void powderBehavior()
    {
        movementPwdr();
        burn();
    }

    /**
     * Movement for type Powder
     */
    private void movementPwdr()
    {
        World myWorld = getWorld();
        Actor d = getOneObjectAtOffset(0, 1, Objects.class);

        if (getY() < myWorld.getHeight()-1 && d == null)
        {
            for (dx = -2; dx <= 2; dx++)
            {
                dy = Greenfoot.getRandomNumber(2);
                int x = getX() + dx;
                int y = getY() + dy;
                Actor obj = getOneObjectAtOffset(dx, dy, Objects.class);
                if (Greenfoot.getRandomNumber(20) == 0 && x >= 0 && x < myWorld.getWidth()-(100/Field.size) 
                && y < myWorld.getHeight() && obj == null)
                    movement();
            }
        }
    }

    /**
     * Behavior for type Spark. Will move downward until it hits the bottom or another Object.
     * Can electrify Objects such as Water, Metal, and Glass.
     */
    private void sparkBehavior()
    {
        movementSpark();
    }

    /**
     * Movement for type Spark
     */
    private void movementSpark()
    {
        World myWorld = getWorld();

        if (getY() < myWorld.getHeight()-1)
        {
            dy = 1;
            dx = Greenfoot.getRandomNumber(3) - 1;
            int newX = getX() + dx;
            Actor obj = getOneObjectAtOffset(dx, dy, Objects.class);
            if (newX >= 0 && newX < myWorld.getWidth()-(100/Field.size))
            {
                if (obj == null)
                    movement();
                else if (((Objects)(obj)).getType() != type)
                    myWorld.removeObject(this);
            }
        } else
            myWorld.removeObject(this);
    }

    /**
     * Basic movement.
     */
    private void movement()
    {
        setLocation(getX()+dx, getY()+dy);
    }

    /**
     * Basic liquid movement.
     */
    private void movementLqd()
    {
        World myWorld = getWorld();
        Actor d = getOneObjectAtOffset(0, 1, Objects.class);
        Actor l = getOneObjectAtOffset(-1, 0, Objects.class);
        Actor r = getOneObjectAtOffset(1, 0, Objects.class);

        if (getY() < myWorld.getHeight()-1 && d == null)
        {
            //Nothing below the Object.
            dy = 1;
            if (Greenfoot.getRandomNumber(5) == 0)
                dx = (Greenfoot.getRandomNumber(2) == 0) ? 1: -1;
        } else {
            //Object is at the bottom of the Field or directly above another Object.
            if (Greenfoot.getRandomNumber(2) == 0)
            {
                if (r == null)
                    dx = 1;
                else if (l == null)
                    dx = -1;
                else
                    return;
            } else {
                if (l == null)
                    dx = -1;
                else if (r == null)
                    dx = 1;
                else
                    return;
            }
        }

        //Final checks to keep Object within boundaries and prevent intersection with other Objects.
        int x = getX()+dx;
        Actor o = getOneObjectAtOffset(dx, dy, Objects.class);
        if (o == null && x >= 0 && x < myWorld.getWidth()-(100/Field.size))
            movement();
        dx = 0;
        dy = 0;
    }
    int movableTypes[] = {1, 2, 5, 9};
    /**
     * Basic gaseous movement. Checks a number of locations above the Object and selects an empty one at random.
     */
    private void movementGas()
    {
        World myWorld = getWorld();

        for (dx = -1; dx <= 1; dx++)
        {
            dy = -Greenfoot.getRandomNumber(2);
            int x = getX()+dx;
            int y = getY()+dy;
            Actor o = getOneObjectAtOffset(dx, dy, Objects.class);
            if (Greenfoot.getRandomNumber(6) == 0 && x >= 0 && x < myWorld.getWidth()-(100/Field.size) && y >= 0)
            {
                if (o == null)
                    movement();
                else {
                    int objType = ((Objects)(o)).getType();
                    for (int i = 0; i < 4; i++)
                    {
                        if (objType == movableTypes[i])
                        {
                            o.setLocation(getX(), getY());
                            movement();
                            return;
                        }
                    }
                }
            }
        }
    }
    public static int totalGroups = 0;
    public int group = 0;
    public int getGroup()
    {
        return group;
    }

    /**
     * Objects which perform heat transfer will form groups based on Object type and physical state.
     * Helps determine contiguous groups of similar Objects for use in heat tranfer calculations.
     */
    public void formGroups()
    {
        if (state == 1 && getY() == getWorld().getHeight()-1)
        {
            //Liquids on the lowest layer of the Field tend to spread out but still count as one group.
            //Checks left and right for other similar Objects and sets group to the lowest group number found.
            int lowestGroup = totalGroups;
            for (int i = -1; i >= -getX(); i--)
            {
                //Checking left
                Actor o = getOneObjectAtOffset(i, 0, Objects.class);
                if (o != null)
                {
                    if (((Objects)(o)).getType() == type && ((Objects)(o)).getState() == state)
                    {
                        int otherGroup = ((Objects)(o)).getGroup();
                        lowestGroup = (otherGroup != 0 && otherGroup < lowestGroup) ? otherGroup: lowestGroup;
                    } else
                        break;
                }
            }
            for (int i = 1; i < getWorld().getWidth()-(100/Field.size)-getX(); i++)
            {
                //Checking right
                Actor o = getOneObjectAtOffset(i, 0, Objects.class);
                if (o != null)
                {
                    if (((Objects)(o)).getType() == type && ((Objects)(o)).getState() == state)
                    {
                        int otherGroup = ((Objects)(o)).getGroup();
                        lowestGroup = (otherGroup != 0 && otherGroup < lowestGroup) ? otherGroup: lowestGroup;
                    } else
                        break;
                }
            }
            //Setting group
            if (lowestGroup != 0)
            {
                if (group == 0)
                    group = lowestGroup;
                else
                    group = (lowestGroup < group) ? lowestGroup: group;
            }
            else if (group == 0)
                group = totalGroups++;
        }
        else if (!getObjectsInRange(1, Objects.class).isEmpty())
        {
            //Sets group number to lowest value of adjacent similar Objects.
            List o = getObjectsInRange(1, Objects.class);
            int sameTypeAdjacent = 0;
            for (int i = 0; i < o.size(); i++)
            {
                if (((Objects)(o.get(i))).getType() == type && ((Objects)(o.get(i))).getState() == state)
                {
                    int otherGroup = ((Objects)(o.get(i))).getGroup();
                    if (group == 0)
                        group = (otherGroup > 0) ? otherGroup: totalGroups++;
                    else
                        group = (otherGroup > 0 && otherGroup < group) ? group = otherGroup: group;
                    sameTypeAdjacent++;
                }
            }
            if (sameTypeAdjacent == 0) group = 0;
        } else
            group = 0;
    }

    int delay = 0;
    int delayTime = 5;
    /**
     * Controls the transfer of heat between certain Objects and physical state changes.
     */
    private void transfer()
    {
        formGroups();

        if (delay > 0)
            delay--;
        else
            delay = delayTime;

        if (!getObjectsInRange(1, Objects.class).isEmpty())
        {
            List o = getObjectsInRange(1, Objects.class);
            int size = o.size();
            if (delay == 0)
            {
                if (electrified)
                    heat++;
                //Detects the number of Objects in a group for use in calculation of heat transfer.
                int groupSize = 1;
                if (group != 0 && !getWorld().getObjects(Objects.class).isEmpty())
                {
                    List obj = getWorld().getObjects(Objects.class);
                    groupSize = 0;
                    for (int i = 0; i < obj.size(); i++)
                        if (((Objects)(obj.get(i))).getGroup() == group)
                            groupSize++;
                }
                //Q represents the net change in heat energy based on the temperatures of adjacent Objects.
                double Q = 0;
                //double highestTemp = 0;
                for (int i = 0; i < size; i++)
                {
                    double tempDif = 0;
                    //Prevented from detecting the temperature of types Wall and Clone.
                    //Their temperatures never change and so they would become infinite heat sinks.
                    if (((Objects)(o.get(i))).getType() != 0 && ((Objects)(o.get(i))).getType() != 4)
                    {
                        tempDif = ((Objects)(o.get(i))).getHeat() - heat;
                        //highestTemp = (((Objects)(o.get(i))).getHeat() > highestTemp) ? ((Objects)(o.get(i))).getHeat(): highestTemp;
                    }
                    Q += (k*tempDif)/groupSize;
                }
                //delT represents the change in an Object's temperature based on Q, mass (m), and specific heat (c)
                double delT = (c > 0) ? (Q / (c*m)): 0;
                
                double roundDelT = Math.abs((int)(delT * 1000) / 1000);
                if (roundDelT <= 0.4 && size < 4 && heat != roomTemp)
                {
                    heat = (heat > roomTemp) ? heat - (4 - size): heat + (4 - size);
                }
                
                //Determines the next physical state the Object is approaching.
                //Currently, Water (type 1) is the only type which can perform heat transfer and be a gas (state 2). 
                int nextState = -1;
                if (delT < 0)
                    nextState = (state > 0) ? state--: 1;
                else if (delT > 0)
                {
                    if (type != 1)
                        nextState = (state == 0) ? 1: 0;
                    else
                        nextState = (state < 2) ? state++: 1;
                }
                //Physical state changes and temperature change.
                if (nextState == 0 && state == 1)
                {
                    //Changing from liquid to solid (state 1 to 0)
                    if (heat + delT > meltingPoint[type])
                        heat = (heat + delT > -273) ? heat + delT: -273;
                    else if (heat + delT > meltingPoint[type] - (int)(meltingPoint[type] / 3))
                    {
                        heat = meltingPoint[type];
                        if (group != 0 && !getWorld().getObjects(Objects.class).isEmpty())
                        {
                            List obj = getWorld().getObjects(Objects.class);
                            int readyStateChange = 0;
                            for (int i = 0; i < obj.size(); i++)
                                if (((Objects)(obj.get(i))).getGroup() == group && ((Objects)(obj.get(i))).getHeat() <= meltingPoint[type] + 15)
                                    readyStateChange++;
                            
                            if (readyStateChange == groupSize)
                                state = 0;
                        }
                    } else {
                        heat = (heat + delT > -273) ? heat + delT: -273;
                        state = 0;
                    }
                }
                else if (nextState == 1 && state == 0)
                {
                    //Changing from solid to liquid (state 0 to 1)
                    if (heat + delT < meltingPoint[type])
                        heat = (heat + delT > -273) ? heat + delT: -273;
                    else if (heat + delT < meltingPoint[type] + (int)(meltingPoint[type] / 3))
                    {
                        heat = meltingPoint[type];
                        if (group != 0 && !getWorld().getObjects(Objects.class).isEmpty())
                        {
                            List obj = getWorld().getObjects(Objects.class);
                            int readyStateChange = 0;
                            for (int i = 0; i < obj.size(); i++)
                                if (((Objects)(obj.get(i))).getGroup() == group && ((Objects)(obj.get(i))).getHeat() >= meltingPoint[type] - 15)
                                    readyStateChange++;

                            if (readyStateChange == groupSize)
                                state = 1;
                        }
                    } else {
                        heat = (heat + delT > -273) ? heat + delT: -273;
                        state = 1;
                    }
                }
                else if (nextState == 2)
                {
                    //Changing from liquid to gas (state 1 to 2)
                    if (heat + delT < 100)
                        heat = (heat + delT > -273) ? heat + delT: -273;
                    else if (heat + delT < 100 + (int)(100 / 3))
                    {
                        heat = 100;
                        if (group != 0 && !getWorld().getObjects(Objects.class).isEmpty())
                        {
                            List obj = getWorld().getObjects(Objects.class);
                            int readyStateChange = 0;
                            for (int i = 0; i < obj.size(); i++)
                                if (((Objects)(obj.get(i))).getGroup() == group && ((Objects)(obj.get(i))).getHeat() >= 100 - 15)
                                    readyStateChange++;

                            if (readyStateChange == groupSize)
                                state = 2;
                        }
                    } else {
                        heat = (heat + delT > -273) ? heat + delT: -273;
                        state = 2;
                    }
                }
                else if (nextState == 1 && state == 2)
                {
                    //Changing from gas to liquid (state 2 to 1)
                    if (heat + delT > 100)
                        heat = (heat + delT > -273) ? heat + delT: -273;
                    else if (heat + delT > 100 - (int)(100 / 3))
                    {
                        heat = 100;
                        if (group != 0 && !getWorld().getObjects(Objects.class).isEmpty())
                        {
                            List obj = getWorld().getObjects(Objects.class);
                            int readyStateChange = 0;
                            for (int i = 0; i < obj.size(); i++)
                                if (((Objects)(obj.get(i))).getGroup() == group && ((Objects)(obj.get(i))).getHeat() <= 100 + 15)
                                    readyStateChange++;

                            if (readyStateChange == groupSize)
                                state = 1;
                        }
                    } else {
                        heat = (heat + delT > -273) ? heat + delT: -273;
                        state = 1;
                    }
                }
                else if ((type == 1 && state == 2 && delT > 0) || 
                        (state == 0 && delT < 0) || (type != 1 && state == 1 && delT > 0))
                    heat = (heat + delT > -273) ? heat + delT: -273;
            }
        }
    }

    /**
     * Used in testing. Calculates the change in temperature for one Object using the same method as above.
     */
    public double calculateStuff()
    {
        int groupSize = 1;
        double Q = 0;
        double delT = 0;
        if (!getObjectsInRange(1, Objects.class).isEmpty())
        {
            List o = getObjectsInRange(1, Objects.class);

            if (group != 0 && !getWorld().getObjects(Objects.class).isEmpty())
            {
                List obj = getWorld().getObjects(Objects.class);
                groupSize = 0;
                for (int i = 0; i < obj.size(); i++)
                    if (((Objects)(obj.get(i))).getGroup() == group)
                        groupSize++;
            }

            for (int i = 0; i < o.size(); i++)
            {
                double tempDif = ((Objects)(o.get(i))).getHeat() - heat;
                Q += (k*tempDif)/groupSize;
            }

            delT = (c > 0) ? (Q / (c*m)): 0;

        }
        //return groupSize;
        //return Q;
        return delT;
    }

    /**
     * Causes Object to be replaced with Fire when it comes in contact with another Object of a certain heat.
     */
    private void burn()
    {
        World myWorld = getWorld();

        if (!getObjectsInRange(1, Objects.class).isEmpty())
        {
            List objects = getObjectsInRange(1, Objects.class);
            for (int i = 0; i < objects.size(); i++)
            {
                if (((Objects)(objects.get(i))).getHeat() >= 250 && ((Objects)(objects.get(i))).getType() != 1
                && Greenfoot.getRandomNumber(10) == 0)
                {
                    type = 3;
                    name = names[type];
                    heat = heats[type];
                    r = ri[type];
                    g = gi[type];
                    b = bi[type];
                    setImage(new GreenfootImage(Field.size, Field.size));
                    GreenfootImage img = getImage();
                    img.setColor(new Color(r, g, b));
                    img.fill();
                }
            }
        }
    }
    int timeElectrified = 10;
    /**
     * Certain Objects will become electrified when in contact with Spark or other electrified Objects.
     * Being electrified will change the Object's appearance for a short time.
     */
    private void electrify()
    {

        if (!getObjectsInRange(1, Objects.class).isEmpty() && !electrified && timeElectrified == 10)
        {
            List objects = getObjectsInRange(1, Objects.class);
            for (int i = 0; i < objects.size(); i++)
                if (((Objects)(objects.get(i))).getElectrified())
                    electrified = true;
        }
        else if (electrified || timeElectrified != 10)
        {
            if (getImage().getColor() != new Color(ri[10], gi[10], bi[10]) && electrified)
            {
                GreenfootImage electric = new GreenfootImage(Field.size, Field.size);
                electric.setColor(new Color(ri[10], gi[10], bi[10]));
                electric.fill();
                setImage(electric);
            }
            if (timeElectrified <= 5)
                electrified = false;
            if (timeElectrified > 0)
                timeElectrified--;
            else {
                timeElectrified = 10;
                //Changing state value to get the Water image to reset.
                state = 3;
                //Changing adjacent value to get the Glass image to reset.
                adjacent = -1;
            }
        }
    }
}
