import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)

/**
 * Enables pause functionality while simulation is running.
 * When paused, all Objects will cease functioning but the Cursor can still be used.
 * 
 * @author Mr. Hedgeapple 
 * @version 1.1
 */
public class PauseButton extends Actor
{
    public static boolean paused = true;
    public static boolean getPaused()
    {
        return paused;
    }

    /**
     * Basic constructor for PauseButton.
     */
    public PauseButton()
    {
        update();
    }

    /**
     * Act - do whatever the PauseButton wants to do. This method is called whenever
     * the 'Act' or 'Run' button gets pressed in the environment.
     */
    public void act() 
    {
        if (Greenfoot.mouseClicked(this))
            update();
    }    

    /**
     * Refreshes the image of PauseButton and flips the boolean paused.
     */
    private void update()
    {
        if (paused)
        {
            setImage(new GreenfootImage(" || ", 40, Color.WHITE, Color.BLACK));
            paused = false;
        } else {
            setImage(new GreenfootImage(" || ", 40, Color.BLACK, Color.WHITE));
            paused = true;
        }
        GreenfootImage img = getImage();
        img.setColor(Color.WHITE);
        img.drawRect(0, 0, img.getWidth()-1, img.getHeight()-1);
    }
}
